@@ -1,29 +1,100 @@
+import contextlib
+import fnmatch
 import logging
 import socket
 import stat
-from contextlib import contextmanager
+import time
 from os.path import expanduser
+from pathlib import Path, PurePosixPath
+from typing import Tuple, List
 
 import paramiko
 
-from antareslauncher.remote_environnement.iconnection import IConnection
+try:
+    # noinspection PyUnresolvedReferences
+    from typing import TypeAlias
+except ImportError:
+    RemotePath = PurePosixPath
+    LocalPath = Path
+else:
+    RemotePath: TypeAlias = PurePosixPath
+    LocalPath: TypeAlias = Path
 
 
-class SshConnection(IConnection):
+class SshConnectionError(Exception):
+    """
+    SSH Connection Error
+    """
+
+
+class InvalidConfigError(SshConnectionError):
+    def __init__(self, config, msg=""):
+        err_msg = f"Invalid configuration error {config}"
+        if msg:
+            err_msg += f": {msg}"
+        super().__init__(err_msg)
+
+
+class ConnectionFailedException(SshConnectionError):
+    def __init__(self, hostname: str, port: int, username: str):
+        msg = (
+            f"Unable to connect to {hostname} on port {port} with username {username}."
+            f" Please ensure that the hostname, port, and username are correct, and"
+            f" that the server is reachable and accepting SSH connections."
+            f" If you are using a password to authenticate, please double-check that"
+            f" the password is correct. If you are using a private key, please ensure"
+            f" that the key is in the correct format and that you have specified"
+            f" the correct path to the key file."
+        )
+        super().__init__(msg)
+
+
+class DownloadMonitor:
+    def __init__(self, total_size: int, msg: str = "", logger=None) -> None:
+        self.total_size = total_size
+        self.msg = msg or "Downloading..."
+        self.logger = logger or logging.getLogger(__name__)
+        self._start_time = time.time()
+        self._size = 0
+        self._progress: int = 0
+
+    def __call__(self, transferred: int, subtotal: int) -> None:
+        if not self.total_size:
+            return
+        self._size += transferred
+        # Avoid emitting too many messages
+        rate = self._size / self.total_size
+        if self._progress != int(rate * 10):
+            self._progress = int(rate * 10)
+            self.logger.info(str(self))
+
+    def __str__(self):
+        rate = self._size / self.total_size
+        if self._size:
+            # Calculate ETA and progress rate
+            # 0        curr_size                   total_size
+            # |----------->|--------------------------->|
+            # 0        duration                    total_duration
+            # 0%       percent                         100%
+            duration = time.time() - self._start_time
+            eta = int(duration * (self.total_size - self._size) / self._size)
+            return f"{self.msg:<20} ETA: {eta}s [{rate:.0%}]"
+        return f"{self.msg:<20} ETA: ??? [{rate:.0%}]"
+
+
+class SshConnection:
     """Class to _connect to remote server"""
 
-    class ConnectionFailedException(Exception):
-        pass
-
     def __init__(self, config: dict = None):
         """
+        Initialize the SSH connection.
 
         Args:
             config: Dictionary containing the "hostname" (name or IP), "username", "port" (default is 22),
             "password" (not compulsory if private_key_file is given), "private_key_file": path to private rsa key
         """
         super(SshConnection, self).__init__()
-        self.logger = logging.getLogger(__name__ + "." + __class__.__name__)
+        self.logger = logging.getLogger(f"{__name__}.{__class__.__name__}")
         self.__client = None
         self.__home_dir = None
         self.timeout = 10
@@ -37,7 +108,11 @@ class SshConnection(IConnection):
             self.logger.info("Loading ssh connection from config dictionary")
             self.__init_from_config(config)
         else:
-            raise IOError
+            error = InvalidConfigError(
+                config, "missing values: 'hostname', 'username', 'password'..."
+            )
+            self.logger.debug(str(error))
+            raise error
         self.initialize_home_dir()
         self.logger.info(
             f"Connection created with host = {self.host} and username = {self.username}"
@@ -71,19 +146,16 @@ class SshConnection(IConnection):
         self.host = config.get("hostname", "")
         self.username = config.get("username", "")
         self.port = config.get("port", 22)
-        self.password = config.get("password", None)
-        key_password = config.get("key_password", None)
-        key_file = config.get("private_key_file", None)
-        if key_file:
-            key_file_path = expanduser(key_file)
+        self.password = config.get("password")
+        key_password = config.get("key_password")
+        if key_file := config.get("private_key_file"):
             self.__initialise_public_key(
-                key_file_name=key_file_path, key_password=key_password
+                key_file_name=key_file, key_password=key_password
             )
         elif self.password is None:
-            self.logger.debug(
-                "self.password is None, no key found, now password was given"
-            )
-            raise ValueError
+            error = InvalidConfigError(config, "missing 'password'")
+            self.logger.debug(str(error))
+            raise error
 
     def initialize_home_dir(self):
         """Initializes self.__home_dir with the home directory retrieved by started "echo $HOME" connecting to the
@@ -101,13 +173,14 @@ class SshConnection(IConnection):
         """
         return self.__home_dir
 
-    @contextmanager
-    def ssh_client(self):
+    @contextlib.contextmanager
+    def ssh_client(self) -> paramiko.SSHClient:
         client = paramiko.SSHClient()
         try:
             try:
                 # Paramiko.SSHClient can be used to make connections to the remote server and transfer files
-                # Parsing an instance of the AutoAddPolicy to set_missing_host_key_policy() changes it to allow any host.
+                # Parsing an instance of the AutoAddPolicy to set_missing_host_key_policy()
+                # changes it to allow any host.
                 client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
                 # Connect to the server
                 if self.private_key:
@@ -130,19 +203,28 @@ class SshConnection(IConnection):
                         allow_agent=False,
                         look_for_keys=False,
                     )
-            except paramiko.AuthenticationException:
+            except paramiko.AuthenticationException as e:
                 self.logger.exception(
                     f"paramiko.AuthenticationException: {paramiko.AuthenticationException}"
                 )
-            except paramiko.SSHException:
+                raise ConnectionFailedException(
+                    self.host, self.port, self.username
+                ) from e
+            except paramiko.SSHException as e:
                 self.logger.exception(f"paramiko.SSHException: {paramiko.SSHException}")
-                raise SshConnection.ConnectionFailedException()
-            except socket.timeout:
+                raise ConnectionFailedException(
+                    self.host, self.port, self.username
+                ) from e
+            except socket.timeout as e:
                 self.logger.exception(f"socket.timeout: {socket.timeout}")
-                raise SshConnection.ConnectionFailedException()
-            except socket.error:
+                raise ConnectionFailedException(
+                    self.host, self.port, self.username
+                ) from e
+            except socket.error as e:
                 self.logger.exception(f"socket.error: {socket.error}")
-                raise SshConnection.ConnectionFailedException()
+                raise ConnectionFailedException(
+                    self.host, self.port, self.username
+                ) from e
 
             yield client
         finally:
@@ -161,7 +243,6 @@ class SshConnection(IConnection):
             error: The standard error of the command
         """
         output = None
-        error = None
         self.logger.info(f"Executing command on remote server: {command}")
         try:
             with self.ssh_client() as client:
@@ -172,7 +253,7 @@ class SshConnection(IConnection):
             error = f"Command timed out: {command}"
         except paramiko.SSHException:
             error = f"Failed to execute {command}"
-        except SshConnection.ConnectionFailedException:
+        except ConnectionFailedException:
             error = f"Failed to connect to remote host and execute {command}"
 
         self.logger.info(f"Command output:\nOutput: {output}\nError: {error}")
@@ -197,13 +278,13 @@ class SshConnection(IConnection):
                 sftp_client.put(src, dst)
                 sftp_client.close()
         except paramiko.SSHException:
-            self.logger.debug("Paramiko SSH Exception")
+            self.logger.debug("Paramiko SSH Exception", exc_info=True)
             result_flag = False
         except IOError:
-            self.logger.debug("IO Error")
+            self.logger.debug("IO Error", exc_info=True)
             result_flag = False
-        except SshConnection.ConnectionFailedException:
-            self.logger.error(f"Failed to connect to remote host")
+        except ConnectionFailedException:
+            self.logger.error("Failed to connect to remote host", exc_info=True)
             result_flag = False
         return result_flag
 
@@ -218,23 +299,104 @@ class SshConnection(IConnection):
         Returns:
             True if the file has been downloaded, False otherwise
         """
+        self.logger.info(f'Downloading remote file "{src}" to directory "{dst}"')
         try:
             with self.ssh_client() as client:
-                self.logger.info(
-                    f'Downloading remote file "{src}" to directory "{dst}"'
-                )
                 sftp_client = client.open_sftp()
                 sftp_client.get(src, dst)
                 sftp_client.close()
                 result_flag = True
         except paramiko.SSHException:
-            self.logger.error("Paramiko SSH Exception")
+            self.logger.error("Paramiko SSH Exception", exc_info=True)
             result_flag = False
-        except SshConnection.ConnectionFailedException:
-            self.logger.error(f"Failed to connect to remote host")
+        except ConnectionFailedException:
+            self.logger.error("Failed to connect to remote host", exc_info=True)
             result_flag = False
         return result_flag
 
+    def download_files(
+        self,
+        src_dir: RemotePath,
+        dst_dir: LocalPath,
+        pattern: str,
+        *patterns: str,
+    ) -> List[LocalPath]:
+        """
+        Download files matching the specified patterns from the remote
+        source directory to the local destination directory,
+        and remove them when the download is finished.
+
+        Args:
+            src_dir: Remote source directory.
+
+            dst_dir: Local destination directory.
+
+            pattern: Unix shell-style wildcards to match the files to download.
+
+            patterns: Additional Unix shell-style wildcards.
+
+        Returns:
+            The paths of the downloaded files on the local filesystem.
+        """
+        try:
+            return self._download_files(src_dir, dst_dir, (pattern,) + patterns)
+        except TimeoutError as exc:
+            self.logger.error(f"Timeout: {exc}", exc_info=True)
+            return []
+        except paramiko.SSHException:
+            self.logger.error("Paramiko SSH Exception", exc_info=True)
+            return []
+        except ConnectionFailedException:
+            self.logger.error("Failed to connect to remote host", exc_info=True)
+            return []
+
+    def _download_files(
+        self,
+        src_dir: RemotePath,
+        dst_dir: LocalPath,
+        patterns: Tuple[str],
+    ) -> List[LocalPath]:
+        """
+        Download files matching the specified patterns from the remote
+        source directory to the local destination directory.
+
+        Args:
+            src_dir: Remote source directory.
+
+            dst_dir: Local destination directory.
+
+            patterns: A tuple of Unix shell-style wildcards to match the files to download.
+
+        Returns:
+            The paths of the downloaded files on the local filesystem.
+        """
+        with self.ssh_client() as client:
+            with contextlib.closing(
+                client.open_sftp()
+            ) as sftp:  # type: paramiko.sftp_client.SFTPClient
+                # Get list of files to download
+                remote_attrs = sftp.listdir_attr(str(src_dir))
+                remote_files = [file_attr.filename for file_attr in remote_attrs]
+                total_size = sum((file_attr.st_size or 0) for file_attr in remote_attrs)
+                files_to_download = [
+                    f
+                    for f in remote_files
+                    if any(fnmatch.fnmatch(f, pattern) for pattern in patterns)
+                ]
+                # Monitor the download progression
+                monitor = DownloadMonitor(total_size, logger=self.logger)
+                # First get all, then delete all (for reentrancy)
+                count = len(files_to_download)
+                for no, filename in enumerate(files_to_download, 1):
+                    monitor.msg = f"Downloading '{filename}' [{no}/{count}]..."
+                    src_path = src_dir.joinpath(filename)
+                    dst_path = dst_dir.joinpath(filename)
+                    sftp.get(str(src_path), str(dst_path), monitor)
+                for filename in files_to_download:
+                    src_path = src_dir.joinpath(filename)
+                    sftp.remove(str(src_path))
+                return [dst_dir.joinpath(filename) for filename in files_to_download]
+
     def check_remote_dir_exists(self, dir_path):
         """Checks if a remote path is a directory
 
@@ -259,10 +421,10 @@ class SshConnection(IConnection):
                 else:
                     raise IOError
         except FileNotFoundError:
-            self.logger.debug("FileNotFoundError")
+            self.logger.debug("FileNotFoundError", exc_info=True)
             result_flag = False
-        except SshConnection.ConnectionFailedException:
-            self.logger.error(f"Failed to connect to remote host")
+        except ConnectionFailedException:
+            self.logger.error("Failed to connect to remote host", exc_info=True)
             result_flag = False
         return result_flag
 
@@ -276,7 +438,7 @@ class SshConnection(IConnection):
             True if file exists and is not empty, False otherwise
 
         Raises:
-            IOError if path exists and it is a directory
+            IOError if path exists, and it is a directory
         """
         result_flag = False
         try:
@@ -286,17 +448,14 @@ class SshConnection(IConnection):
                 sftp_stat = sftp_client.stat(file_path)
                 sftp_client.close()
                 if stat.S_ISREG(sftp_stat.st_mode):
-                    if sftp_stat.st_size > 0:
-                        result_flag = True
-                    else:
-                        result_flag = False
+                    result_flag = sftp_stat.st_size > 0
                 else:
-                    raise IOError
+                    raise IOError(f"Not a regular file: '{file_path}'")
         except FileNotFoundError:
-            self.logger.debug("FileNotFoundError")
+            self.logger.debug("FileNotFoundError", exc_info=True)
             result_flag = False
-        except SshConnection.ConnectionFailedException:
-            self.logger.error(f"Failed to connect to remote host")
+        except ConnectionFailedException:
+            self.logger.error("Failed to connect to remote host", exc_info=True)
             result_flag = False
         return result_flag
 
@@ -318,10 +477,7 @@ class SshConnection(IConnection):
                 try:
                     self.logger.info(f"Checking if remote directory {dir_path} exists")
                     sftp_stat = sftp_client.stat(dir_path)
-                    if stat.S_ISDIR(sftp_stat.st_mode):
-                        result_flag = True
-                    else:
-                        result_flag = False
+                    result_flag = stat.S_ISDIR(sftp_stat.st_mode)
                 except FileNotFoundError:
                     self.logger.info(f"Creating remote directory {dir_path}")
                     sftp_client.mkdir(dir_path)
@@ -329,10 +485,10 @@ class SshConnection(IConnection):
                 finally:
                     sftp_client.close()
         except paramiko.SSHException:
-            self.logger.debug("Paramiko SSHException")
+            self.logger.debug("Paramiko SSHException", exc_info=True)
             result_flag = False
-        except SshConnection.ConnectionFailedException:
-            self.logger.error(f"Failed to connect to remote host")
+        except ConnectionFailedException:
+            self.logger.error("Failed to connect to remote host", exc_info=True)
             result_flag = False
         return result_flag
 
@@ -346,30 +502,26 @@ class SshConnection(IConnection):
             True if file is successfully removed, False otherwise
 
         Raises:
-            IOError if path exists and it is a directory
+            IOError if path exists, and it is a directory
         """
-        result_flag: bool = False
         try:
             with self.ssh_client() as client:
                 sftp_client = client.open_sftp()
                 try:
                     self.logger.info(f"Removing remote file {file_path}")
                     sftp_stat = sftp_client.stat(file_path)
-                    if stat.S_ISREG(sftp_stat.st_mode):
-                        try:
-                            sftp_client.remove(file_path)
-                        except IOError:
-                            pass
-                        result_flag = True
-                    else:
-                        raise IOError
+                    if not stat.S_ISREG(sftp_stat.st_mode):
+                        raise IOError(f"Not a regular file: '{file_path}'")
+                    with contextlib.suppress(IOError):
+                        sftp_client.remove(file_path)
+                    result_flag = True
                 except FileNotFoundError:
-                    self.logger.debug("FileNotFound nothing to remove")
+                    self.logger.debug("FileNotFound nothing to remove", exc_info=True)
                     result_flag = True
                 finally:
                     sftp_client.close()
-        except SshConnection.ConnectionFailedException:
-            self.logger.error(f"Failed to connect to remote host")
+        except ConnectionFailedException:
+            self.logger.error("Failed to connect to remote host", exc_info=True)
             result_flag = False
         return result_flag
 
@@ -383,37 +535,33 @@ class SshConnection(IConnection):
             True if the directory is successfully removed, False otherwise
 
         Raises:
-            IOError if path exists and it is a file
+            IOError if path exists, and it is a file
         """
-        result_flag: bool = False
         try:
             with self.ssh_client() as client:
                 sftp_client = client.open_sftp()
                 try:
                     self.logger.info(f"Removing remote directory {dir_path}")
                     sftp_stat = sftp_client.stat(dir_path)
-                    if stat.S_ISDIR(sftp_stat.st_mode):
-                        try:
-                            sftp_client.rmdir(dir_path)
-                        except IOError:
-                            pass
-                        result_flag = True
-                    else:
-                        raise IOError
+                    if not stat.S_ISDIR(sftp_stat.st_mode):
+                        raise IOError(f"Not a directory: '{dir_path}'")
+                    with contextlib.suppress(IOError):
+                        sftp_client.rmdir(dir_path)
+                    result_flag = True
                 except FileNotFoundError:
-                    self.logger.debug("DirNotFound nothing to remove")
+                    self.logger.debug("DirNotFound nothing to remove", exc_info=True)
                     result_flag = True
                 finally:
                     sftp_client.close()
-        except SshConnection.ConnectionFailedException:
-            self.logger.error(f"Failed to connect to remote host")
+        except ConnectionFailedException:
+            self.logger.error("Failed to connect to remote host", exc_info=True)
             result_flag = False
         return result_flag
 
     def test_connection(self):
         try:
-            with self.ssh_client() as client:
+            with self.ssh_client():
                 return True
-        except SshConnection.ConnectionFailedException:
-            self.logger.error(f"Failed to connect to remote host")
+        except ConnectionFailedException:
+            self.logger.error("Failed to connect to remote host", exc_info=True)
             return False
